local merkVehicleUtil
local merkSettingsUtil
local version = 1
local minorVersion = 2

require "stringutil"

-- base functions
local function fileExists(fileName)
	local file = io.open(fileName, "r")
	if file then
		file:close()
		return true
	end
	return false
end

local next = next
local function tableEmpty(t)
	return next(t) == nil
end

local function tableEqual(t1, t2)
	if type(t1) ~= "table" or type(t2) ~= "table" then
		return false
	end
	local isEqual = true
	local checkedKeys = {}
	for k, v in pairs(t1) do
		if (t2[k] == nil) or (type(t2[k]) ~= type(v)) then
			isEqual = false
			break
		elseif type(v) == "table" then
			isEqual = tableEqual(v, t2[k])
			if not isEqual then
				break
			end
		elseif v ~= t2[k] then
			isEqual = false
			break
		end
		checkedKeys[k] = true
	end
	if isEqual then
		for k, v in pairs(t2) do
			if not checkedKeys[k] then
				if (t1[k] == nil) or (type(t1[k]) ~= type(v)) then
					isEqual = false
					break
				elseif type(v) == "table" then
					isEqual = tableEqual(v, t1[k])
					if not isEqual then
						break
					end
				elseif v ~= t1[k] then
					isEqual = false
					break
				end
			end
		end
	end
	return isEqual
end

local function tableCopy(input)
	if type(input) == "table" then
		local result = {}
		for k, v in pairs(input) do
			if type(v) == "table" then
				result[k] = tableCopy(v)
			else
				result[k] = v
			end
		end
		return result
	else
		return input
	end
end

local function getModPath()
	local info
	local i = 1
	repeat
		info = debug.getinfo(i,'S')
		i = i+1
	until info == nil
	
	local modPath = ""
	if i > 1 then
		info = debug.getinfo(i-2,'S')
		if info then
			if string.ends(info.source, 'mod.lua') then
				modPath = string.gsub(info.source, "@(.*/)mod[.]lua", "%1")
			elseif string.ends(info.source, '.mdl') then
				modPath = string.gsub(info.source, "@(.*/)res/models/model/.+[.]mdl", "%1")
			elseif string.ends(info.source, '.con') then
				modPath = string.gsub(info.source, "@(.*/)res/construction/.+[.]con", "%1")
			elseif string.ends(info.source, '.msh') then
				modPath = string.gsub(info.source, "@(.*/)res/models/mesh/.+[.]msh", "%1")
			elseif string.ends(info.source, '.grp') then
				modPath = string.gsub(info.source, "@(.*/)res/models/group/.+[.]grp", "%1")
			elseif string.ends(info.source, '.mtl') then
				modPath = string.gsub(info.source, "@(.*/)res/models/material/.+[.]mtl", "%1")
			elseif string.ends(info.source, '.ani') then
				modPath = string.gsub(info.source, "@(.*/)res/models/animation/.+[.]ani", "%1")
			elseif string.find(info.source, "/workshop/content/446800/") then
				modPath = string.gsub(info.source, "@(.*/workshop/content/446800/[^/]*/).*", "%1")
			elseif string.find(info.source, "/staging_area/") then
				modPath = string.gsub(info.source, "@(.*/staging_area/[^/]*/).*", "%1")
			elseif string.find(info.source, "/mods/") then
				modPath = string.gsub(info.source, "@(.*/mods/[^/]*/).*", "%1")
			end
		end
	end
	
	return modPath
end

local function clamp(value, min, max)
	if type(value) == "number" then
		return math.min(math.max(value, min), max)
	else
		return 0
	end
end

local function writeBytes(x, numBytes)
	local size = 2
	if numBytes then
		size = numBytes
	end
	local result = ""
	local value = x or 0
	for i = 1, size do
		result = result..string.char(value%256)
		value=(value-value%256)/256
	end
    return result
end

local defaultBytes = string.char(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

local function readBytes(file, numBytes)
	if numBytes == 1 then
		return string.byte(file:read(1) or "0")
	else
		local size = 2
		if numBytes then
			size = numBytes
		end
		local byte_data = file:read(size) or defaultBytes
		local result = 0
		for i = 0, size-1 do
			result = result + 256 ^ i * string.byte(byte_data, i+1)
		end
		return result
	end
end

-- Repaints
local function addMaterial(id, file, materials)
	if type(merkVehicleUtil) == "table" and type(merkVehicleUtil[id]) == "table" and type(materials) == "table" then
		if type(merkVehicleUtil[id].meshData) == "table" and type(merkVehicleUtil[id].meshData[file]) == "table" then
			local mesh = merkVehicleUtil[id].meshData[file]
			local matConfig = {}
			if type(mesh.materials) == "table" then
				for i, subMesh in ipairs(mesh.materials) do
					local matConf = 0
					for k, material in ipairs(materials) do
						if type(subMesh) == "table" and type(material) == "table" and material[1] and material[2] then
							for j, mat in ipairs(subMesh) do
								if mat == material[1] and mat ~= material[2] then
									subMesh[#subMesh+1] = material[2]
									matConf = #subMesh-1
								end
								if mat == material[2] and (j < #subMesh or matConf == 0) then
									-- material already exists
									if matConf > 0 then
										subMesh[#subMesh] = nil
										matConf = j-1
									end
									break
								end
							end
						end
					end
					matConfig[#matConfig+1] = matConf
				end
			end
			
			local matConfigId = #mesh.matConfigs+1
			if type(mesh.matConfigs) == "table" and #matConfig > 0 then
				for i, matConf in ipairs(mesh.matConfigs) do
					if tableEqual(matConf, matConfig) then
						matConfigId = i
						break
					end
				end
				if matConfigId > #mesh.matConfigs then
					mesh.matConfigs[matConfigId] = matConfig
				end
				return matConfigId-1
			end
		end
	end
	return nil
end

local function getGroupMatConfig(id, file, matConfigs)
	if type(merkVehicleUtil) == "table" and type(merkVehicleUtil[id]) == "table" then
		if type(merkVehicleUtil[id].groupData) == "table" and type(merkVehicleUtil[id].groupData[file]) == "table" then
			local group = merkVehicleUtil[id].groupData[file]
			local matConfig = {}
			
			if type(group.meshes) == "table" and type(matConfigs) == "table" then
				for i, mesh in ipairs(group.meshes) do
					if matConfigs[mesh] then
						matConfig[#matConfig+1] = matConfigs[mesh]
					else
						matConfig[#matConfig+1] = 0
					end
				end
				
				if type(group.matConfigs) == "table" then
					local matConfigID = #group.matConfigs+1
					for i, matConf in ipairs(group.matConfigs) do
						if tableEqual(matConf, matConfig) then
							matConfigID = i
							break
						end
					end
					if matConfigID > #group.matConfigs then
						group.matConfigs[matConfigID] = matConfig
					end
					return matConfigID-1
				end
			end
		end
	end
	return nil
end

local function addRepaint(id, repaintId, materials, files)
	if type(id) == "string" and type(repaintId) == "string" and type(merkVehicleUtil) == "table" then
		if type(merkVehicleUtil.repaintModels) ~= "table" then
			merkVehicleUtil.repaintModels = {}
		end
		local repaintModels = merkVehicleUtil.repaintModels
		
		if type(merkVehicleUtil[id]) == "table" and merkVehicleUtil[id].created then
			-- make sure "meshData" and "groupData" are tables
			local meshData = {}
			if type(merkVehicleUtil[id].meshData) == "table" then
				meshData = merkVehicleUtil[id].meshData
			end
			local groupData = {}
			if type(merkVehicleUtil[id].groupData) == "table" then
				groupData = merkVehicleUtil[id].groupData
			end
			
			local meshFiles = {}
			local groupFiles = {}
			if type(files) == "table" and #files > 0 then
				for i, file in ipairs(files) do
					if meshData[file] then
						meshFiles[#meshFiles+1] = file
					elseif groupData[file] then
						groupFiles[#groupFiles+1] = file
					end
				end
			else
				for file, data in pairs(meshData) do
					meshFiles[#meshFiles+1] = file
				end
				for file, data in pairs(groupData) do
					groupFiles[#groupFiles+1] = file
				end
			end
			
			local matConfigs = {}
			for i, file in ipairs(meshFiles) do
				matConfigs[file] = addMaterial(id, file, materials)
			end
			
			for i, file in ipairs(groupFiles) do
				matConfigs[file] = getGroupMatConfig(id, file, matConfigs)
			end
			
			if type(merkVehicleUtil[id].repaints) ~= "table" then
				merkVehicleUtil[id].repaints = {}
			end
			
			merkVehicleUtil[id].repaints[repaintId] = matConfigs
		else
			-- main mod not loaded, store data for later use
			if type(merkVehicleUtil[id]) ~= "table" then
				merkVehicleUtil[id] = {}
			end
			if type(merkVehicleUtil[id].pendingRepaints) ~= "table" then
				merkVehicleUtil[id].pendingRepaints = {}
			end
			merkVehicleUtil[id].pendingRepaints[#merkVehicleUtil[id].pendingRepaints+1] = {
				id = repaintId,
				materials = materials,
				files = files
			}
		end
		
		if type(files) == "table" and type(files.models) == "table" then
			for i, model in ipairs(files.models) do
				repaintModels[model] = { mod_id = id, repaint_id = repaintId }
			end
		end
	else
		-- wrong parameters
		if type(id) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #1 to 'addRepaint' (string expected, got %s)", type(id)))
		end
		if type(repaintId) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #2 to 'addRepaint' (string expected, got %s)", type(repaintId)))
		end
	end
end

local function repaintAvailable(id, repaintId)
	if type(id) == "string" and type(repaintId) == "string" and type(merkVehicleUtil) == "table" and type(merkVehicleUtil[id]) == "table" then
		if type(merkVehicleUtil[id].repaints) == "table"  and type(merkVehicleUtil[id].repaints[repaintId]) == "table" then
			return true
		end
	end
	
	return false
end

local function getRepaint(id, repaintId, data)
	if type(id) == "string" and type(repaintId) == "string" and type(merkVehicleUtil) == "table" and type(merkVehicleUtil[id]) == "table" and type(merkVehicleUtil[id].repaints) == "table" then
		if type(merkVehicleUtil[id].repaints[repaintId]) == "table" then
			local matConfigs = merkVehicleUtil[id].repaints[repaintId]
		
			if type(data) == "table" and type(data.lods) == "table" then
				for i, lod in ipairs(data.lods) do
					local matConfig = {}
					if type(lod) == "table" then
						for j, child in ipairs(lod.children) do
							if matConfigs[child.id] then
								matConfig[#matConfig+1] = matConfigs[child.id]
							elseif lod.matConfigs[1][j] then
								matConfig[#matConfig+1] = lod.matConfigs[1][j]
							else
								matConfig[#matConfig+1] = 0
							end
						end
					else
						print("[ModUtil|Vehicles] Missing key \"lod.children\".")
					end
					lod.matConfigs[1] = matConfig
				end
			end
	
			return matConfigs
		end
	else
		-- wrong parameters
		if type(id) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #1 to 'getRepaint' (string expected, got %s)", type(id)))
		end
		if type(repaintId) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #2 to 'getRepaint' (string expected, got %s)", type(repaintId)))
		end
	end
	
	if type(data) == "table" then
		data.isHiddenRepaint = true
	end
	
	return {}
end

local function getMeshData(id, fileName, data)
	if type(id) == "string" and type(fileName) == "string" and type(merkVehicleUtil) == "table" and type(merkVehicleUtil[id]) == "table" and type(merkVehicleUtil[id].meshData) == "table" then
		local meshData = merkVehicleUtil[id].meshData[fileName]
		
		if type(meshData) == "table" and data then
			if meshData.matConfigs then
				data.matConfigs = meshData.matConfigs
			end
			if type(meshData.materials) == "table" and type(data.subMeshes) == "table" then
				for i, subMesh in ipairs(data.subMeshes) do
					if meshData.materials[i] then
						subMesh.materials = meshData.materials[i]
					end
				end
			end
		end
		
		return meshData
	else
		-- wrong parameters
		if type(id) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #1 to 'getMeshData' (string expected, got %s)", type(id)))
		end
		if type(fileName) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #2 to 'getMeshData' (string expected, got %s)", type(fileName)))
		end
	end
end

local function getGroupData(id, fileName, data)
	if type(id) == "string" and type(fileName) == "string" and type(merkVehicleUtil) == "table" and type(merkVehicleUtil[id]) == "table" and type(merkVehicleUtil[id].groupData) == "table" then
		local groupData = merkVehicleUtil[id].groupData[fileName]
		
		if type(groupData) == "table" and data then
			if groupData.matConfigs then
				data.matConfigs = groupData.matConfigs
			end
		end
		
		return groupData
	else
		-- wrong parameters
		if type(id) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #1 to 'getGroupData' (string expected, got %s)", type(id)))
		end
		if type(fileName) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #2 to 'getGroupData' (string expected, got %s)", type(fileName)))
		end
	end
end

-- Metadata
local function changeMetadataTable(metadata, key, value, table_add)
	if type(metadata) == "table" then
		if metadata[key] then
			if type(value.value) == "table" then
				local new_table
				if type(metadata[key]) ~= "table" then
					if value.action == "REPLACE" or value.action == "ADD_REPLACE" then
						metadata[key] = {}
						new_table = not value.emptyTable
					else
						return
					end
				end
				for k, v in pairs(value.value) do
					changeMetadataTable(metadata[key], k, v, new_table)
				end
			else
				if value.action == "REPLACE" or value.action == "ADD_REPLACE" then
					metadata[key] = value.value
				elseif value.action == "FACTOR" then
					if type(metadata[key]) == "number" and type(value.value) == "number" then
						metadata[key] = metadata[key] * value.value
					end
				end
			end
		elseif value.action == "ADD_REPLACE" or value.action == "ADD" or table_add then
			if type(value.value) == "table" and not value.emptyTable then
				metadata[key] = {}
				for k, v in pairs(value.value) do
					changeMetadataTable(metadata[key], k, v, table_add)
				end
			else
				metadata[key] = value.value
			end			
		end
	end
end

local function changeMetadata(id, fileName, data, fileFilter)
	if type(fileName) == "string" and type(merkVehicleUtil) == "table" then
		local metadata = {}
		if fileFilter then
			if type(merkVehicleUtil.metadataFF) == "table" and type(merkVehicleUtil.metadataFF[fileName]) == "table" then
				metadata = merkVehicleUtil.metadataFF[fileName]
			end
		else
			if type(merkVehicleUtil.metadata) == "table" and type(merkVehicleUtil.metadata[fileName]) == "table" then
				metadata = merkVehicleUtil.metadata[fileName]
			end
		end

		if not tableEmpty(metadata) and (not metadata.done or getModPath() ~= "res/scripts/") and type(data) == "table" and type(data.metadata) == "table" then
			for key, value in pairs(metadata) do
				if key ~= "done" then
					changeMetadataTable(data.metadata, key, value)
				end
			end
			if getModPath() ~= "res/scripts/" then
				metadata.done = true
			end
		end
	else
		-- wrong parameters
		if type(fileName) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #2 to 'changeMetadata' (string expected, got %s)", type(fileName)))
		end
	end
end

local actionList = {
	FACTOR = {
		REPLACE = "REPLACE",
		ADD = "ADD_REPLACE",
		ADD_REPLACE = "ADD_REPLACE",
	},
	REPLACE = {
		FACTOR = "REPLACE",
		ADD = "ADD_REPLACE",
		ADD_REPLACE = "ADD_REPLACE",
	},
	ADD = {
		FACTOR = "ADD_REPLACE",
		REPLACE = "ADD_REPLACE",
		ADD_REPLACE = "ADD_REPLACE",
	},
	ADD_REPLACE = {
		FACTOR = "ADD_REPLACE",
		REPLACE = "ADD_REPLACE",
		ADD = "ADD_REPLACE",
	},
}

local function setMetadataTable(metadataCache, key, value, priority, action, empty)
	if metadataCache[key] == nil or (type(metadataCache[key]) == "table" and type(metadataCache[key].value) ~= "table" and metadataCache[key].priority < priority and not empty) then
		metadataCache[key] = {
			value = {},
			priority = priority,
			action = action,
		}
		if type(value) == "table" then
			metadataCache[key].emptyTable = empty
			for k, v in pairs(value) do
				setMetadataTable(metadataCache[key].value, k, v, priority, action, empty)
			end
		else
			metadataCache[key].value = value
		end
		return action, priority
	elseif type(metadataCache[key]) == "table" and type(metadataCache[key].value) ~= "table" and metadataCache[key].priority >= priority then
		return metadataCache[key].action or "ADD_REPLACE", metadataCache[key].priority
	elseif type(metadataCache[key]) == "table" and type(metadataCache[key].value) == "table" then
		if type(value) == "table" then
			local act, prio
			for k, v in pairs(value) do
				act, prio = setMetadataTable(metadataCache[key].value, k, v, priority, action)
				metadataCache[key].action = actionList[act][metadataCache[key].action or "ADD_REPLACE"]
				metadataCache[key].priority = math.max(prio, metadataCache[key].priority)
			end
			return metadataCache[key].action, metadataCache[key].priority
		else
			if metadataCache[key].priority < priority then
				metadataCache[key] = {
					value = value,
					priority = priority,
					action = action
				}
				return action, priority
			else
				return metadataCache[key].action or "ADD_REPLACE", metadataCache[key].priority
			end
		end
	end
	
	return "ADD_REPLACE", 0
end

local function setMetadata(file_name, meta_data, prio, act)
	if type(merkVehicleUtil) == "table" then
		local fileName = file_name
		local metadata = meta_data
		local priority = prio or 0
		local action = act or "ADD_REPLACE"
		
		if type(meta_data) == "string" and type(prio) == "table" then
			fileName = meta_data
			metadata = prio
			priority = act or 0
			action = "ADD_REPLACE"
		end
		
		if type(fileName) == "string" and type(metadata) == "table" and type(priority) == "number" and type(action) == "string" then
			if type(merkVehicleUtil.metadata) ~= "table" then
				merkVehicleUtil.metadata = {}
			end
			if type(merkVehicleUtil.metadata[fileName]) ~= "table" then
				merkVehicleUtil.metadata[fileName] = {}
			end
			
			if type(merkVehicleUtil.metadataFF) ~= "table" then
				merkVehicleUtil.metadataFF = {}
			end
			if type(merkVehicleUtil.metadataFF[fileName]) ~= "table" then
				merkVehicleUtil.metadataFF[fileName] = {}
			end
			
			local metadataCache
			
			if type(metadata) == "table" then
				for key, value in pairs(metadata) do
					if string.find(key, "[.]") then
						local new_value = {}
						local new_key = string.match(key, "([^.]+)[.]")
						local current_value = new_value
						for k in string.gmatch(key, "[.]([^.]+)[.]") do
							current_value[key] = {}
						end
						
						if new_key == "availability" or new_key == "description" then
							metadataCache = merkVehicleUtil.metadataFF[fileName]
						else
							metadataCache = merkVehicleUtil.metadata[fileName]
						end
						
						setMetadataTable(metadataCache, new_key, new_value, priority, action, true)
						current_value[string.gsub(key, "([^.]+)[.]", "")] = value
						setMetadataTable(metadataCache, new_key, new_value, priority, action)
					else
						if key == "availability" or key == "description" then
							metadataCache = merkVehicleUtil.metadataFF[fileName]
						else
							metadataCache = merkVehicleUtil.metadata[fileName]
						end
						
						setMetadataTable(metadataCache, key, value, priority, action)
					end
				end
			end
			
			-- delete empty tables
			if tableEmpty(merkVehicleUtil.metadata[fileName]) then
				merkVehicleUtil.metadata[fileName] = nil
				if tableEmpty(merkVehicleUtil.metadata) then
					merkVehicleUtil.metadata = nil
				end
			end
			if tableEmpty(merkVehicleUtil.metadataFF[fileName]) then
				merkVehicleUtil.metadataFF[fileName] = nil
				if tableEmpty(merkVehicleUtil.metadataFF) then
					merkVehicleUtil.metadataFF = nil
				end
			end
		else
			-- wrong parameters
			if type(fileName) ~= "string" then
				print(string.format("[ModUtil|Vehicles] bad argument #1 to 'setMetadata' (string expected, got %s)", type(fileName)))
			end
			if type(metadata) ~= "table" then
				print(string.format("[ModUtil|Vehicles] bad argument #2 to 'setMetadata' (table expected, got %s)", type(metadata)))
			end
			if type(priority) ~= "number" then
				print(string.format("[ModUtil|Vehicles] bad argument #3 to 'setMetadata' (number expected, got %s)", type(priority)))
			end
			if type(action) ~= "string" then
				print(string.format("[ModUtil|Vehicles] bad argument #4 to 'setMetadata' (string expected, got %s)", type(action)))
			end
		end
	end
end

local function modelModifierFunction(fileName, data)
	if type(merkVehicleUtil) == "table" then
		local pattern = "models/model"
		local i = string.find(fileName, pattern)
		if i then
			local modelId = string.sub(fileName, i+#pattern+1)
			
			if type(merkVehicleUtil.repaintModels) == "table" then
				if type(merkVehicleUtil.repaintModels[modelId]) == "table" then
					local repaintData = merkVehicleUtil.repaintModels[modelId]
					getRepaint(repaintData.mod_id, repaintData.repaint_id, data)
				end
			end
			
			if type(merkVehicleUtil.metadata) == "table" then
				changeMetadata("", modelId, data)
			end
		end
	end
	
	return data
end

local function modelModifier(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= 1 then
		return merk_vehicleUtil[version].modelModifierFunction(fileName, data)
	end
	
	return data
end

-- Object filter
local function modelFilter(fileName, data)
	local pattern = "models/model"

	if type(merkVehicleUtil) == "table" then
		local i = string.find(fileName, pattern)
		if i then
			local modelId = string.sub(fileName, i+#pattern+1)

			-- use filter as modifer (so that it is executed before all other modifiers). Works only for availability and description
			if type(merkVehicleUtil.metadataFF) == "table" then
				changeMetadata("", modelId, data, true)
			end
			
			if type(merkVehicleUtil.repaintModels) == "table" then
				if type(merkVehicleUtil.repaintModels[modelId]) == "table" then
					local repaintData = merkVehicleUtil.repaintModels[modelId]
					if not repaintAvailable(repaintData.mod_id, repaintData.repaint_id) then
						return false
					end
				end
			end
			-- modifier end
			
			if type(merkVehicleUtil.hiddenObjects) == "table" and merkVehicleUtil.hiddenObjects[modelId] then
				return false
			end
		end
	end
	
	return true
end

local function objectFilter(fileName, data, pattern)
	if type(merkVehicleUtil) == "table" and type(merkVehicleUtil.hiddenObjects) == "table" then
		local i = string.find(fileName, pattern)
		if i then
			local objectId = string.sub(fileName, i+#pattern+1)
			if merkVehicleUtil.hiddenObjects[objectId] then
				return false
			end
		end
	end
	
	return true
end

local function filterModels(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].modelFilter(fileName, data)
	end
	
	return true
end

local function filterConstructions(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].objectFilter(fileName, data, "construction/")
	end
	
	return true
end

local function filterBridges(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].objectFilter(fileName, data, "config/bridge/")
	end
	
	return true
end

local function filterTunnels(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].objectFilter(fileName, data, "config/tunnel/")
	end
	
	return true
end

local function filterStreets(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].objectFilter(fileName, data, "config/street/")
	end
	
	return true
end

local function filterTracks(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].objectFilter(fileName, data, "config/track/")
	end
	
	return true
end

local function filterMultipleUnits(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].objectFilter(fileName, data, "config/multiple_units/")
	end
	
	return true
end

local function filterRailroadCrossings(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].objectFilter(fileName, data, "config/railroad_crossing/")
	end
	
	return true
end

local function filterGroundTextures(fileName, data)
	if merk_vehicleUtil and merk_vehicleUtil[version] and merk_vehicleUtil[version].minorVersion >= minorVersion then
		return merk_vehicleUtil[version].objectFilter(fileName, data, "config/auto_ground_tex")
	end
	
	return true
end

local function hideObject(objectId)
	if type(objectId) == "string" and type(merkVehicleUtil) == "table" then
		if type(merkVehicleUtil.hiddenObjects) ~= "table" then
			merkVehicleUtil.hiddenObjects = {}
		end
		
		merkVehicleUtil.hiddenObjects[objectId] = true
	else
		-- wrong parameters
		if type(objectId) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #1 to 'hideObject' (string expected, got %s)", type(objectId)))
		end
	end
end

local function readFiles(files, options)
	local meshData = {}
	local groupData = {}
	local dataBackup = data

	local modPath = getModPath()
	
	if fileExists(modPath.."mod.lua") then
		local repaintFileName = modPath.."repaintFiles.lua"
		if type(options) ~= "table" or (options.useRepaintFile ~= false and options.devMode ~= true) then
			local repaintFile = loadfile(repaintFileName)
			if repaintFile then
				return repaintFile()
			end
		end
		
		if type(options) == "table" and (options.loadMeshes or (options.loadMeshes == nil and options.devMode)) then
			for i, file in ipairs(files) do
				if type(file) == "string" then
					if string.ends(file, ".msh") then
						local meshFile = loadfile(modPath.."res/models/mesh/"..file)
						if type(meshFile) == "function" then
							meshFile()
							if type(data) == "function" then
								local mesh = data()
								local mData
								
								if type(mesh) == "table" and type(mesh.matConfigs) == "table" and type(mesh.subMeshes) == "table" then
									mData = {
										materials = {},
									}
									mData.matConfigs = mesh.matConfigs
									for j, subMesh in ipairs(mesh.subMeshes) do
										if subMesh.materials then
											mData.materials[#mData.materials+1] = subMesh.materials
										else
											print(string.format("[ModUtil|Vehicles] Missing field \"submesh.material\" in mesh %q!"), file)
										end
									end
								end
								
								meshData[file] = mData
							end
						end
					elseif string.ends(file, ".grp") then
						local groupFile = loadfile(modPath.."res/models/group/"..file)
						if type(groupFile) == "function" then
							groupFile()
							if type(data) == "function" then
								local group = data()
								local gData
								
								if type(group) == "table" and type(group.matConfigs) == "table" and type(group.children) == "table" then
									gData = {
										meshes = {},
									}
									gData.matConfigs = group.matConfigs
									for j, child in ipairs(group.children) do
										if child.id then
											gData.meshes[#gData.meshes+1] = child.id
										else
											print(string.format("[ModUtil|Vehicles] Missing field \"child.id\" in group %q!"), file)
										end
									end
								end
								
								groupData[file] = gData
							end
						end
					end
				end
			end
			
			data = dataBackup
			
			if options.createRepaintFile or (options.createRepaintFile == nil and options.devMode) then
				local repaintFile = io.open(repaintFileName, "w")
				if repaintFile then
					require "serialize"
					repaintFile:write("local "..serializeStr({ meshData = meshData, groupData = groupData }))
					repaintFile:write("local result = data()\nreturn result.meshData, result.groupData")
					repaintFile:close()
				else
					print("[ModUtil|Vehicles] Creation of repaint file failed")
				end
			end
		end
		
	else
		print("[ModUtil|Vehicles] Could not read files. Maybe the function was called from a wrong file.")
	end
	
	return meshData, groupData
end

local function changeFiles(files, id)
	local modPath = getModPath()
	if type(id) == "string" and type(files) == "table" then
		if id ~= "" then
			for i, file in ipairs(files) do
				if type(file) == "string" then
					local filename = ""
					local fileType = ""
					if string.ends(file, ".msh") then
						filename = modPath.."res/models/mesh/"..file
						fileType = "Mesh"
					elseif string.ends(file, ".grp") then
						filename = modPath.."res/models/group/"..file
						fileType = "Group"
					end
					local dataFile = io.open(filename, "r")
					local text = ""
					if dataFile then
						text = dataFile:read("*a")
						dataFile:close()
						
						if not string.match(text, 'merk_vehicleUtil%[1%].get'..fileType..'Data') then
							text = string.gsub(text, 'function%sdata%(%)%s*return(.*)end', 'local result =%1')
							
							text = text..'\nif merk_vehicleUtil and merk_vehicleUtil[1] then\n'
							text = text..string.format('\tmerk_vehicleUtil[1].get'..fileType..'Data("%s", "%s", result)\n', repaintId, file)
							text = text..'end\n\nfunction data()\n\treturn result\nend'
						end
					end
					
					dataFile = io.open(filename, "w")
					if dataFile then
						dataFile:write(text)
						dataFile:close()
					end						
				end
			end
		end
	else
		-- wrong parameters
		if type(files) ~= "table" then
			print(string.format("[ModUtil|Vehicles] bad argument #1 to 'changeFiles' (table expected, got %s)", type(files)))
		end
		if type(id) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #2 to 'changeFiles' (string expected, got %s)", type(id)))
		end
	end
end

local function createEnvironment(id, files, meshes, groups, options)
	local meshData = {}
	local groupData = {}
	
	if type(id) == "string" then
		if type(files) == "table" and #files > 0 then
			meshData, groupData = readFiles(files, options)
		else
			if type(meshes) == "table" then meshData = meshes end
			if type(groups) == "table" then groupData = groups end
		end
		
		if game or (options and options.modelViewer) then
			if options and options.modelViewer then
				if merkVehicleUtil == nil then
					merkVehicleUtil = {}
				end
			else
				if type(game.merkVehicleUtil) ~= "table" then
					game.merkVehicleUtil = {}
				end
				if type(game.merkVehicleUtil[version]) ~= "table" then
					game.merkVehicleUtil[version] = {}
				end
				merkVehicleUtil = game.merkVehicleUtil[version]
			end
			
			if type(merkVehicleUtil[id]) == "table" and merkVehicleUtil[id].created then
				print(string.format("[ModUtil|Vehicles] Creation of repaint environment for id %q failed, the id is already registered.", id))
			else
				if type(merkVehicleUtil[id]) ~= "table" then
					merkVehicleUtil[id] = {}
				end
				
				merkVehicleUtil[id].created = true
				
				merkVehicleUtil[id].meshData = meshData
				merkVehicleUtil[id].groupData = groupData
				
				if type(merkVehicleUtil[id].pendingRepaints) == "table" and #merkVehicleUtil[id].pendingRepaints > 0 then
					pendingRepaints = merkVehicleUtil[id].pendingRepaints
					merkVehicleUtil[id].pendingRepaints = nil
					for i, repaint in ipairs(pendingRepaints) do
						addRepaint(id, repaint.id, repaint.materials, repaint.files)
					end
				end
				
				return {
					addRepaint = function(repaintId, materials, files)
						return addRepaint(id, repaintId, materials, files)
					end,
					getRepaint = function(repaintId, data)
						return getRepaint(id, repaintId, data)
					end,
					getMeshData = function(fileName)
						return getMeshData(id, fileName)
					end,
					getGroupData = function(fileName)
						return getGroupData(id, fileName)
					end,
					setMetadata = setMetadata,
					changeMetadata = function(fileName, data)
						return changeMetadata(id, fileName, data)
					end,
				}
			end
		else
			print(string.format("[ModUtil|Vehicles] Creation of repaint environment for id %q failed. It seems you're trying to create the environment outside the \"runFn\".", id))
		end
	else
		-- wrong parameters
		if type(id) ~= "string" then
			print(string.format("[ModUtil|Vehicles] bad argument #1 to 'changeFiles' (string expected, got %s)", type(id)))
		end
	end
end

local function modelViewerRepaint(id, repaintId, materials, files, data)
	createEnvironment(id, {""}, nil, nil, {modelViewer = true})
	addRepaint(id, repaintId, materials, files)
	getRepaint(id, repaintId, data)
end

local function initialiseVehicleUtil(addFilters)
	if game then
		if type(game.merkVehicleUtil) ~= "table" then
			game.merkVehicleUtil = {}
		end
		if type(game.merkVehicleUtil[version]) ~= "table" then
			game.merkVehicleUtil[version] = {}
		end
		merkVehicleUtil = game.merkVehicleUtil[version]
	end
	
	
	if addFilters ~= false then
		if addFileFilter then
			addFileFilter("model/vehicle", filterModels)
			addFileFilter("model/tree", filterModels)
			addFileFilter("model/industry", filterModels)
			addFileFilter("model/other", filterModels)
			
			addFileFilter("construction", filterConstructions)
			addFileFilter("bridge", filterBridges)
			addFileFilter("tunnel", filterTunnels)
			addFileFilter("street", filterStreets)
			addFileFilter("track", filterTracks)		
			addFileFilter("multipleUnit", filterMultipleUnits)
			addFileFilter("railroadCrossing", filterRailroadCrossings)
			addFileFilter("autoGroundTex", filterGroundTextures)
		end
		if addModifier then
			addModifier("loadModel", modelModifier)
		end
	end

	merk_vehicleUtil[version] = {
		createEnvironment = createEnvironment,
		addRepaint = addRepaint,
		getRepaint = getRepaint,
		getMeshData = getMeshData,
		getGroupData = getGroupData,
		modelViewerRepaint = modelViewerRepaint,
		changeFiles = changeFiles,
		setMetadata = setMetadata,
		changeMetadata = changeMetadata,
		hideObject = hideObject,
		minorVersion = minorVersion,
		modelFilter = modelFilter,
		objectFilter = objectFilter,
		modelModifierFunction = modelModifierFunction,
	}
end

if not merk_vehicleUtil or not merk_vehicleUtil[version] then

	if type(merk_vehicleUtil) ~= "table" then
		merk_vehicleUtil = {}
	end
	
	initialiseVehicleUtil()
else
	if merk_vehicleUtil[version].minorVersion < minorVersion then
		initialiseVehicleUtil(merk_vehicleUtil[version].minorVersion < 1)
	end
end

-- settings
local function readUserSettings(id)
	if type(merkSettingsUtil) == "table" and type(merkSettingsUtil[id]) == "table" then
		local settings = merkSettingsUtil[id].values
		
		local modPath = getModPath()
		
		if fileExists(modPath.."mod.lua") then
			local settingsFile = loadfile(modPath.."settings.lua")
			local loadedSettings = {}
			
			if type(settingsFile) == "function" then
				loadedSettings = settingsFile()
			else
				print(string.format("[ModUtil|UserSettings] Could not read settings file %q.", modPath.."mod.lua"))
			end
			
			if type(loadedSettings) == "table" then
				for option_key, option in pairs(merkSettingsUtil[id].definitions) do
					if type(loadedSettings[option_key]) == option.type then
						local value = loadedSettings[option_key]
						if option.type == "number" then
							if option.min and value < option.min then
								value = option.min
							end
							if option.max and value > option.max then
								value = option.max
							end
						end
						settings[option_key] = value
					end
				end
			else
				print(string.format("[ModUtil|UserSettings] Settings file %q contains invalid data.", modPath.."mod.lua"))
			end
		end
	end
end

local function createUserSettings(id, options)
	if type(id) == "string" and type(options) == "table" then
		if type(merkSettingsUtil) == "table" then
			if merkSettingsUtil[id] then
				print(string.format("[ModUtil|UserSettings] Previous settings for id %q are overwritten.", id))
			end
			
			merkSettingsUtil[id] = {
				definitions = options,
				values = {},
			}
			
			for option_key, option in pairs(options) do
				if option.type and option.name then
					if option.default ~= nil then
						merkSettingsUtil[id].values[option_key] = option.default
					else
						if option.type == "boolean" then
							merkSettingsUtil[id].values[option_key] = false
						elseif option.type == "number" then
							merkSettingsUtil[id].values[option_key] = 0
						elseif option.type == "string" then
							merkSettingsUtil[id].values[option_key] = ""
						elseif option.type == "table" then
							merkSettingsUtil[id].values[option_key] = {}
						end
					end
				end
			end
			
			readUserSettings(id)
		end
	else
		-- wrong parameters
		if type(id) ~= "string" then
			print(string.format("[ModUtil|UserSettings] bad argument #1 to 'createUserSettings' (string expected, got %s)", type(id)))
		end
		if type(options) ~= "table" then
			print(string.format("[ModUtil|UserSettings] bad argument #2 to 'createUserSettings' (table expected, got %s)", type(options)))
		end
	end
end

local function getUserSettings(id)
	local settings = {}
	
	if type(id) == "string" then
		if type(merkSettingsUtil) == "table" and type(merkSettingsUtil[id]) == "table" then
			settings = merkSettingsUtil[id].values
		end
	else
		-- wrong parameter
		if type(id) ~= "string" then
			print(string.format("[ModUtil|UserSettings] bad argument #1 to 'getUserSettings' (string expected, got %s)", type(id)))
		end
	end
	
	return settings
end

local function initialiseSettingsUtil()
	if game then
		if type(game.merkSettingsUtil) ~= "table" then
			game.merkSettingsUtil = {}
		end
		if type(game.merkSettingsUtil[version]) ~= "table" then
			game.merkSettingsUtil[version] = {}
		end
		merkSettingsUtil = game.merkSettingsUtil[version]
	end

	merk_SettingsUtil[version] = {
		create = createUserSettings,
		get = getUserSettings,
		minorVersion = minorVersion,
	}
end

if not merk_SettingsUtil or not merk_SettingsUtil[version] then
	if type(merk_SettingsUtil) ~= "table" then
		merk_SettingsUtil = {}
	end

	initialiseSettingsUtil()
else
	if merk_SettingsUtil[version].minorVersion < minorVersion then
		initialiseSettingsUtil()
	end
end

-- textures
local function createTargaHeader(width, height, has_alpha)
	local bit_depth = 24
	if has_alpha then
		bit_depth = 32
	end
	
	local header = string.char(0,0,2)..
	writeBytes(0)..
	writeBytes(0)..

	writeBytes(bit_depth, 1)..
	
	writeBytes(0)..
	writeBytes(0)..
	
	-- Breite/Höhe
	writeBytes(width)..
	writeBytes(height)..
	
	writeBytes(bit_depth, 1)..
	writeBytes(0, 1)
	
	return header
end

local function getCharData(charImage, fontData)
	local char_file = io.open(charImage..".tpff", "rb")
	local result = fontData or {}
	if char_file then
		local file_version = string.byte(char_file:read(1))
		if file_version == 1 then
			local num_sizes = string.byte(char_file:read(1))
			for i = 1, num_sizes do
				local fontSize = {}
				if result[i] then
					fontSize = result[i]
					local cache = string.byte(char_file:read(1))
				else
					fontSize.height = string.byte(char_file:read(1))
				end
				local numChars = readBytes(char_file, 4)
				
				for j = 1, numChars do
					local key = readBytes(char_file, 4)
					local char_data = {}
					char_data.width = string.byte(char_file:read(1))
					char_data.height = string.byte(char_file:read(1))
					char_data.mask = {}
					
					for n = 1, char_data.height do
						char_data.mask[n] = {}
						for m = 1, char_data.width do
							char_data.mask[n][m] = string.byte(char_file:read(1))
						end
					end
					
					fontSize[key] = char_data
				end

				if not result[i] and fontSize.height > 0 then
					result[#result+1] = fontSize
				end
			end
		else
			print("[ModUtil|Textures] unsupported font-file version")
		end
		
		char_file:close()
	elseif not fontData then
		print(string.format("[ModUtil|Textures] font-file %q not found"), charImage..".tpff")
	end
	
	if #result > 0 and not fontData then
		-- only load additional files if loading of base file was succesfull
		for i = 1, 9 do
			getCharData(charImage..i, result)
		end
	else
		result = {{height = 1}}
	end
	
	return result
end

local function getUnicodeChars(text)
	local chars = {}
	local i = 1
	while i <= #text do
		local char_index = string.byte(text, i)
		if char_index < 128 then
			chars[#chars+1] = char_index
		elseif char_index >= 194 then
			local bytes = {}
			i = i+1
			bytes[1] = char_index-192
			bytes[2] = string.byte(text, i)-128
			if char_index >= 224 then
				i = i+1
				bytes[1] = char_index-224
				bytes[3] = string.byte(text, i)-128
				if char_index >= 240 then
					i = i+1
					bytes[1] = char_index-240
					bytes[4] = string.byte(text, i)-128
				end
			end
			
			local unicode_char = 0
			for j, char_byte in ipairs(bytes) do
				unicode_char = unicode_char*64 + char_byte
			end
			
			chars[#chars+1] = unicode_char
		end
		i = i+1
	end
	return chars
end

local function createTexts(width, height, texts, font_filename, options)
	if type(width) == "number" and type(height) == "number" and type(texts) == "table" and type(font_filename) == "string" then
		local has_alpha = false
		local text_blocks = {}
		local chars = {}
		
		if texts.text then
			text_blocks[1] = tableCopy(texts)
		else
			text_blocks = tableCopy(texts)
		end
		
		for i, text_block in ipairs(text_blocks) do
			if text_block.text then
				if not text_block.text_color then
					text_block.text_color = {r = 0, g = 0, b = 0, a = 255}
				elseif not text_block.text_color.a then
					text_block.text_color.a = 255
				end
				
				if not text_block.background_color then
					text_block.background_color = {r = 255, g = 255, b = 255, a = 255}
				elseif not text_block.background_color.a then
					text_block.background_color.a = 255
				end
				
				text_block.align = text_block.align or 1
				
				has_alpha = has_alpha or text_block.background_color.a < 255

				local block_chars = getUnicodeChars(text_block.text)
				for j = 1, #block_chars do
					chars[#chars+1] = {id = block_chars[j], block = i}
				end
			else
				text_blocks[i] = nil
			end
		end
		
		local pixels = {}
			
		for i = 1, height do
			pixels[i] = {}
			for j = 1, width do
				pixels[i][j] = {
					r = 255,
					g = 255,
					b = 255,
				}
				if has_alpha then
					pixels[i][j].a = 255
				end
			end
		end
		
		local breaking_whitespace = {
			[9] = true,
			[10] = true,
			[11] = true,
			[12] = true,
			[13] = true,
			[32] = true,
		}
		
		local font_data = getCharData(font_filename)
		
		local textLines = {}
		local split_indices, line_lengths, line_heights
		local sizeIndex = #font_data
		local forced_break = false
		
		-- split text into words (unbreakable blocks)
		local words = {
			{start = 1, stop = 1}
		}
		for i = 1, #chars do
			if breaking_whitespace[chars[i].id] and i > 1 and i < #chars then
				local word_index = #words
				if not(words[#words].start == i and #words > 1 and words[#words-1].forced_break and chars[i].id ~= 10) then
					if words[#words].start == i and #words > 1 and chars[i].id ~= 10 then
						word_index = #words-1
					else
						if chars[i].id == 10 then
							words[#words].forced_break = true
							forced_break = true
						end
					end
					words[word_index].stop = i-1
					words[word_index+1] = {
						start = i+1,
					}
				end
			end
		end
		words[#words].stop = #chars
		
		for n, fontSize in ipairs(font_data) do
			local max_height = 0
			
			-- get additional word data (height, length)
			for w = 1, #words do
				words[w].height, words[w].length = fontSize.height, 0
				for i = words[w].start, words[w].stop do
					if fontSize[chars[i].id] then
						words[w].length = words[w].length + fontSize[chars[i].id].width
						words[w].height = math.max(words[w].height, fontSize[chars[i].id].height)
					end
				end
				words[w].sep_length = 0
				local sep_end = #chars
				if w < #words then
					sep_end = words[w+1].start-1
				end
				for i = words[w].stop+1, sep_end do
					if fontSize[chars[i].id] then
						words[w].sep_length = words[w].sep_length + fontSize[chars[i].id].width
					end
				end
				max_height = math.max(max_height, words[w].height)
			end
			
			split_indices = {}
			line_lengths = {}
			line_heights = {}
			if max_height <= height then
				local text_length = 0
				for i = 1, #chars do
					if fontSize[chars[i].id] then
						text_length = text_length + fontSize[chars[i].id].width
					end
				end
				
				local length_factor = text_length/width
				
				if length_factor <= 1 and not forced_break then
					line_lengths = {text_length}
					line_heights = {max_height}
				elseif fontSize.height*math.ceil(length_factor) <= height or n == #font_data then
					-- try to break text
					
					local line = 1
					local multiline = true
					
					if length_factor <= 2 and (not options or options.shortFirstLine ~= false) then
						-- different handling for two line texts
						multiline = false
						for i = 1, 2 do
							line_lengths[i] = 0
						end
						local first_length_min = math.max(math.ceil(width * (length_factor - 1)), 0)
						line_heights = {0}
						
						for i = 1, #words do
							if line_lengths[line] > width or (i > 1 and words[i-1].forced_break and (line > 1 or line_lengths[line] < first_length_min)) then
								multiline = true
								line_lengths = {}
								line_heights = {}
								split_indices = {}
								line = 1
								break
							elseif line == 1 and (line_lengths[line] > first_length_min or (first_length_min > 0 and line_lengths[line] >= first_length_min)) then
								if line_heights[line] + max_height <= height then
									line = 2
									line_heights[2] = 0
									split_indices[1] = words[i-1].stop
								else
									-- get height of remaining words
									line_heights[2] = 0
									for j = i+1, #words do
										line_heights[2] = math.max(line_heights[2], words[i].height)
									end
									if line_heights[line] + line_heights[2] <= height then
										line = 2
										split_indices[1] = words[i-1].stop
									else
										line_heights[2] = 0
									end
								end
							end
							
							if i > 1 and line_lengths[line] > 0 then
								line_lengths[line] = line_lengths[line] + words[i-1].sep_length
							end
							line_lengths[line] = line_lengths[line] + words[i].length
							line_heights[line] = math.max(line_heights[line], words[i].height)
						end
					end
					if line <= #line_lengths and line_lengths[line] > width then
						multiline = true
						line_lengths = {}
						line_heights = {}
						split_indices = {}
						line = 1
					end
					
					if multiline then
						line_lengths = {0}
						line_heights = {0}
						for i = 1, #words do
							local new_length = line_lengths[line]
							
							if i > 1 and new_length > 0 then
								new_length = new_length + words[i-1].sep_length
							end
							new_length = new_length + words[i].length
							
							if new_length > width or (i > 1 and words[i-1].forced_break) then
								if i > 1 then
									split_indices[line] = words[i-1].stop
									line = line + 1
									line_lengths[line] = words[i].length
									line_heights[line] = words[i].height
								elseif n ~= #font_data then
									break
								elseif i == 1 then
									line_lengths[line] = new_length
									line_heights[line] = math.max(line_heights[line], words[i].height)
								end
							else
								line_lengths[line] = new_length
								line_heights[line] = math.max(line_heights[line], words[i].height)
							end
						end
					end
					if line_lengths[line] + words[#words].sep_length <= width then
						line_lengths[line] = line_lengths[line] + words[#words].sep_length
					end
				end
				
				local valid_lengths = true
				if n ~= #font_data then
					-- check if combined height for all lines is lower than height
					local full_height = 0
					for i = 1, #line_heights do
						full_height = full_height + line_heights[i]
					end
					if full_height > height then
						valid_lengths = false
					end
				end
				
				for i = 1, #line_lengths do
					valid_lengths = valid_lengths and (line_lengths[i] < width)
				end
				
				if n == #font_data then
					valid_lengths = true
				end
				
				if #split_indices == #line_lengths - 1 and valid_lengths then
					sizeIndex = n
					break
				end
			end
		end
		
		local fontSize = font_data[sizeIndex]
		local lastSize
		if sizeIndex > 1 then
			lastSize = font_data[sizeIndex-1].height
		end
					
		local yoffset = math.max(math.floor((height-#line_lengths*fontSize.height) / (2*(#line_lengths))), 0)
		local bigger_first_line = false
		
		if #line_lengths == 2 and #split_indices == 1 and lastSize and (not options or options.bigFirstLine ~= false) then
			local new_height = line_heights[2] + lastSize
			if new_height <= height then
				-- try to use bigger size for first line
				local length = 0
				local line_height = lastSize
				for i = 1, split_indices[1] do
					if font_data[sizeIndex-1][chars[i].id] then
						length = length + font_data[sizeIndex-1][chars[i].id].width
						line_height = math.max(line_height, font_data[sizeIndex-1][chars[i].id].height)
					end
				end
				new_height = line_heights[2] + line_height
				if length < width and new_height <= height then
					line_lengths[1] = length
					bigger_first_line = true
					yoffset = math.max(math.floor((height-new_height) / (2*(#line_lengths))), 0)
				end
			end
		end
		
		for i = 1, #line_lengths do
			textLines[i] = {}
			if i == 1 and bigger_first_line and sizeIndex > 1 then
				textLines[i].size = sizeIndex-1
			else
				textLines[i].size = sizeIndex
			end
			if i == 1 then
				textLines[i].start = 1
				if bigger_first_line then
					textLines[i].height = height - (2*yoffset + line_heights[2])
				elseif #line_lengths == 1 then
					textLines[i].height = height
				else
					textLines[i].height = 2*yoffset + line_heights[i]
				end
				textLines[i].yoffset = 0
			else
				textLines[i].start = split_indices[i-1]+2
				textLines[i].yoffset = textLines[i-1].yoffset + textLines[i-1].height
				if i == #line_lengths then
					textLines[i].height = height - textLines[i].yoffset
				else
					textLines[i].height = 2*yoffset + line_heights[i]
				end
			end
			if i < #line_lengths then
				textLines[i].length = split_indices[i]
			else
				textLines[i].length = #chars
			end
			textLines[i].yoffset_text = yoffset
			textLines[i].xoffset = math.max(math.floor((width-line_lengths[i]) / 2), 0)

			textLines[i].blocks = {}
			local last_block
			local font = font_data[textLines[i].size]
			for j = textLines[i].start, textLines[i].length do
				local block
				if #textLines[i].blocks > 0 then
					block = textLines[i].blocks[#textLines[i].blocks]
				end
				if last_block and (chars[j].block == last_block) then
					-- add char to existing block
					block.stop = j
					if font[chars[j].id] then
						block.length = block.length + font[chars[j].id].width
					end
				else
					local new_block = {}
					new_block.start = j
					new_block.stop = j
					new_block.length = 0
					new_block.index = chars[j].block
					if font[chars[j].id] then
						new_block.length = font[chars[j].id].width
					end
					new_block.align = 1
					if text_blocks[new_block.index] then
						new_block.align = text_blocks[new_block.index].align
						if block and block.align > new_block.align then
							new_block.align = block.align
						end
					end
					
					if new_block.align == 0 then
						new_block.xoffset = 0
					elseif new_block.align == 1 then
						if not block or block.align == 0 then
							new_block.xoffset = textLines[i].xoffset
						else
							new_block.xoffset = 0
						end
					else
						if block and block.align == 1 then
							new_block.xoffset = 0
							block.length = block.length + (width - line_lengths[i] - textLines[i].xoffset)
						elseif not block or block.align == 0 then
							new_block.xoffset = width - line_lengths[i]
						else
							new_block.xoffset = 0
						end
					end
					new_block.length = new_block.length + new_block.xoffset
					
					textLines[i].blocks[#textLines[i].blocks+1] = new_block
					last_block = chars[j].block
				end
			end
			
			-- empty lines created by \n
			if #textLines[i].blocks == 0 then
				textLines[i].blocks = {{
					start = textLines[i].start,
					stop = textLines[i].length,
					length = width,
					index = chars[textLines[i].start].block,
					align = 0,
					xoffset = 0,
				}}
			else
				-- always extend last block to the end of line
				local block = textLines[i].blocks[#textLines[i].blocks]
				if block.align == 0 then
					block.length = block.length + math.max((width - line_lengths[i]), 0)
				elseif block.align == 1 then
					block.length = block.length + math.max((width - line_lengths[i] - textLines[i].xoffset), 0)
				end
			end
		end
		
		for l = 1, #textLines do
			local yoffset = textLines[l].yoffset
			local xoffset = 0
			local text_offset = textLines[l].yoffset_text
			local size = textLines[l].size
			
			if yoffset < 0 then
				break
			end
			
			for b, block in ipairs(textLines[l].blocks) do
				if text_blocks[block.index] then
					local start = xoffset
					local text_color = text_blocks[block.index].text_color
					local back_color = text_blocks[block.index].background_color
					
					for j = 1, block.xoffset do
						for k = 1, textLines[l].height do
							pixels[yoffset+k][xoffset+j].r = back_color.r
							pixels[yoffset+k][xoffset+j].g = back_color.g
							pixels[yoffset+k][xoffset+j].b = back_color.b
							if has_alpha then
								pixels[yoffset+k][xoffset+j].a = back_color.a
							end
						end
					end
					xoffset = xoffset + block.xoffset
					
					for i = block.start, block.stop do
						local char_data = font_data[size][chars[i].id]
						
						if char_data then						
							if xoffset + char_data.width > width then
								break
							end
							
							for j = 1, char_data.width do
								for k = 1, textLines[l].height do
									local alpha
									if k > text_offset and k <= text_offset + char_data.height then
										alpha = char_data.mask[k-text_offset][j]/255
									else
										alpha = 0
									end
									
									if options and options.matrix then
										if alpha < 0.5 then
											alpha = 0
										else
											alpha = 1
										end
									end
									pixels[yoffset+k][xoffset+j].r = math.floor(back_color.r*(1-alpha) + text_color.r*alpha)
									pixels[yoffset+k][xoffset+j].g = math.floor(back_color.g*(1-alpha) + text_color.g*alpha)
									pixels[yoffset+k][xoffset+j].b = math.floor(back_color.b*(1-alpha) + text_color.b*alpha)
									if has_alpha then
										pixels[yoffset+k][xoffset+j].a = math.floor((1.0-((1.0-back_color.a/255)*(1.0-alpha)))*255)
									end
								end
							end
							xoffset = xoffset + char_data.width
						end
					end
					
					for j = xoffset+1, math.min(start+block.length, width) do
						for k = 1, textLines[l].height do
							pixels[yoffset+k][j].r = back_color.r
							pixels[yoffset+k][j].g = back_color.g
							pixels[yoffset+k][j].b = back_color.b
							if has_alpha then
								pixels[yoffset+k][j].a = back_color.a
							end
						end
					end
					xoffset = start+block.length
				end
			end
		end
		
		return pixels, has_alpha
	else
		-- wrong parameters
		if type(width) ~= "number" then
			print(string.format("[ModUtil|Textures] bad argument #1 to 'textImage' (number expected, got %s)", type(width)))
		end
		if type(height) ~= "number" then
			print(string.format("[ModUtil|Textures] bad argument #2 to 'textImage' (number expected, got %s)", type(height)))
		end
		if type(texts) ~= "table" then
			print(string.format("[ModUtil|Textures] bad argument #3 to 'textImage' (table expected, got %s)", type(texts)))
		end
		if type(font_filename) ~= "string" then
			print(string.format("[ModUtil|Textures] bad argument #4 to 'textImage' (string expected, got %s)", type(font_filename)))
		end
		
		return {}, false
	end
end

local function createTGAFile(filename, width, height, pixels, has_alpha)
	if type(filename) == "string" and type(width) == "number" and type(height) == "number" and type(pixels) == "table" then
		if #pixels >= height and #pixels[1] >= width then
			local textmap = io.open(filename..".tga", "wb")
			if textmap then
				textmap:setvbuf("full")
				textmap:write(createTargaHeader(width, height, has_alpha))
				for i = height, 1, -1 do
					if type(pixels[i]) == "table" and #pixels[i] >= width then
						for j = 1, width do
							if type(pixels[i][j]) == "table" then
								-- BGR
								textmap:write(writeBytes(pixels[i][j].b, 1)..writeBytes(pixels[i][j].g, 1)..writeBytes(pixels[i][j].r, 1))
								if has_alpha then
									textmap:write(writeBytes(pixels[i][j].a, 1))
								end
							else
								textmap:write(writeBytes(0, 3))
								if has_alpha then
									textmap:write(writeBytes(255, 1))
								end
							end
						end
					end
				end
				textmap:flush()
				textmap:close()
			else
				print(string.format("[ModUtil|Textures] Could not create image %q. Failed to open or create file.", filename))
			end
		else
			print(string.format("[ModUtil|Textures] Could not create image %q. Not enough pixel data for the given dimensions.", filename))
		end
	else
		-- wrong parameters
		if type(filename) ~= "string" then
			print(string.format("[ModUtil|Textures] bad argument #1 to 'createTGAFile' (string expected, got %s)", type(filename)))
		end
		if type(width) ~= "number" then
			print(string.format("[ModUtil|Textures] bad argument #2 to 'createTGAFile' (number expected, got %s)", type(width)))
		end
		if type(height) ~= "number" then
			print(string.format("[ModUtil|Textures] bad argument #3 to 'createTGAFile' (number expected, got %s)", type(height)))
		end
		if type(pixels) ~= "table" then
			print(string.format("[ModUtil|Textures] bad argument #5 to 'createTGAFile' (table expected, got %s)", type(pixels)))
		end
	end
end

local function textImage(filename, width, height, texts, font_filename, options)
	if type(filename) == "string" and type(width) == "number" and type(height) == "number" and type(texts) == "table" and type(font_filename) == "string" then
		local script_filename = filename..".lua"
		local currentText = {w = width, h = height, t = {}, f = font_filename, o = {}}
		local newText = {w = width, h = height, t = texts, f = font_filename, o = options}

		script_file = loadfile(script_filename)
		if script_file then
			currentText = script_file()
		end
		
		if not tableEqual(currentText, newText) then
			createTGAFile(filename, width, height, createTexts(width, height, texts, font_filename, options))
			local scriptFile = io.open(script_filename, "w")
			if scriptFile then
				if type(serializeStr) ~= "function" then
					require "serialize"
				end
				scriptFile:write("local "..serializeStr(newText))
				scriptFile:write("return data()")
				scriptFile:close()
			end
		end
	else
		-- wrong parameters
		if type(filename) ~= "string" then
			print(string.format("[ModUtil|Textures] bad argument #1 to 'textImage' (string expected, got %s)", type(filename)))
		end
		if type(width) ~= "number" then
			print(string.format("[ModUtil|Textures] bad argument #2 to 'textImage' (number expected, got %s)", type(width)))
		end
		if type(height) ~= "number" then
			print(string.format("[ModUtil|Textures] bad argument #3 to 'textImage' (number expected, got %s)", type(height)))
		end
		if type(texts) ~= "table" then
			print(string.format("[ModUtil|Textures] bad argument #4 to 'textImage' (table expected, got %s)", type(texts)))
		end
		if type(font_filename) ~= "string" then
			print(string.format("[ModUtil|Textures] bad argument #5 to 'textImage' (string expected, got %s)", type(font_filename)))
		end
	end
end

local function rgba(r, g, b, a)
	local color = {}
	local components = {"r", "g", "b", "a"}
	local defaults = {0, 0, 0, 255}
	if type(r) == "string" then
		if string.sub(r, 1, 1) == "#" then
			if #r == 4 then
				for i = 1, 3 do
					local value = tonumber(string.sub(r, i+1, i+1), 16)
					if value then
						color[components[i]] = value*17
					else
						color[components[i]] = defaults[i]
					end
				end
				color["a"] = 255
			else
				for i = 1, 4 do
					color[components[i]] = defaults[i]
					if i*2+1 <= #r then
						local value = tonumber(string.sub(r, i*2, i*2+1), 16)
						if value then
							color[components[i]] = value
						end
					end
				end
			end
		else
			for i, c in ipairs(components) do
				color[c] = defaults[i]
			end
		end
	else
		color = {r = r, g = g, b = b, a = a}
		if type(r) == "table" then
			color = {r = r.r, g = r.g, b = r.b, a = r.a}
		end
		for i, c in ipairs(components) do
			if type(color[c]) == "number" then
				if color[c] > 0 and color[c] < 1 then
					color[c] = math.ceil(color[c]*255)
				elseif color[c] <= 255 then
					color[c] = math.floor(color[c])
				else
					color[c] = defaults[i]
				end
			else
				color[c] = defaults[i]
			end
		end
	end
	
	return color
end

local function initialiseTextureUtil()
	merk_TextureUtil[version] = {
		createTexts = createTexts,
		createTGAFile = createTGAFile,
		textImage = textImage,
		rgba = rgba,
		minorVersion = minorVersion,
	}
end

if not merk_TextureUtil or not merk_TextureUtil[version] then
	if type(merk_TextureUtil) ~= "table" then
		merk_TextureUtil = {}
	end

	initialiseTextureUtil()
else
	if merk_TextureUtil[version].minorVersion < minorVersion then
		initialiseTextureUtil()
	end
end

-- return value
local vehicles = tableCopy(merk_vehicleUtil[version])
vehicles.objectFilter = nil
vehicles.modelFilter = nil
vehicles.modelModifierFunction = nil

local mod_util = {
	vehicles = vehicles,
	userSettings = tableCopy(merk_SettingsUtil[version]),
	initialize = function (id, settings, files, meshes, groups, options)
			merk_SettingsUtil[version].create(id, settings)
			return merk_vehicleUtil[version].createEnvironment(id, files, meshes, groups, options)
		end,
	minorVersion = minorVersion,
	misc = {
		getModPath = getModPath,
		clamp = clamp,
	},
	textures = tableCopy(merk_TextureUtil[version]),
}

if not merk_modutil or not merk_modutil[version] then
	if type(merk_modutil) ~= "table" then
		merk_modutil = {}
	end
	merk_modutil[version] = mod_util
else
	if merk_modutil[version].minorVersion < minorVersion then
		merk_modutil[version] = mod_util
		package.loaded["merk_modutil_1"] = nil
	end
end

return merk_modutil[version]
